from .constants import RecordCode


# ABSTRACTION

class Record:
    def __init__(self, code, fields, rows=None):
        self.code = code
        self.fields = fields
        self.rows = rows or []


class Bai2Model:
    code = None

    def as_string(self):
        return '\n'.join([
            '{0},{1}'.format(row[0].value, row[1])
            for row in self.rows
        ])


class Bai2SingleModel(Bai2Model):
    def __init__(self, rows=None, **fields):
        self.rows = rows or []
        for name, value in fields.items():
            setattr(self, name, value)


class Bai2SectionModel(Bai2Model):
    def __init__(self, header=None, trailer=None, children=None):
        self.header = header
        self.trailer = trailer
        self.children = children

    def update_totals(self):
        pass

    @property
    def rows(self):
        if not hasattr(self, '_rows'):
            rows = self.header.rows
            for child in self.children:
                rows += child.rows
            rows += self.trailer.rows
            self._rows = rows
        return self._rows


# IMPLEMENTATION

class Bai2File(Bai2SectionModel):
    def __init__(self, header=None, trailer=None, children=None):
        super().__init__(header=header or Bai2FileHeader([]),
                         trailer=trailer or Bai2FileTrailer([]),
                         children=children or [])

    def update_totals(self):
        file_control_total = 0
        for group in self.children:
            file_control_total += group.trailer.group_control_total

        self.trailer.file_control_total = file_control_total
        self.trailer.number_of_groups = len(self.children)


class Bai2FileHeader(Bai2SingleModel):
    code = RecordCode.file_header

    def __init__(
        self,
        rows=None,
        sender_id=None,
        receiver_id=None,
        creation_date=None,
        creation_time=None,
        file_id=None,
        physical_record_length=None,
        block_size=None,
        version_number=2
    ):
        super().__init__(
            rows, sender_id=sender_id, receiver_id=receiver_id,
            creation_date=creation_date, creation_time=creation_time,
            file_id=file_id, physical_record_length=physical_record_length,
            block_size=block_size, version_number=version_number
        )


class Bai2FileTrailer(Bai2SingleModel):
    code = RecordCode.file_trailer

    def __init__(
        self,
        rows=None,
        file_control_total=None,
        number_of_groups=None,
        number_of_records=None
    ):
        super().__init__(
            rows, file_control_total=file_control_total,
            number_of_groups=number_of_groups, number_of_records=number_of_records
        )


class Group(Bai2SectionModel):
    def __init__(self, header=None, trailer=None, children=None):
        super().__init__(header=header or GroupHeader([]),
                         trailer=trailer or GroupTrailer([]),
                         children=children or [])

    def update_totals(self):
        group_control_total = 0
        for account in self.children:
            group_control_total += account.trailer.account_control_total

        self.trailer.group_control_total = group_control_total
        self.trailer.number_of_accounts = len(self.children)


class GroupHeader(Bai2SingleModel):
    code = RecordCode.group_header

    def __init__(
        self,
        rows=None,
        ultimate_receiver_id=None,
        originator_id=None,
        group_status=None,
        as_of_date=None,
        as_of_time=None,
        currency=None,
        as_of_date_modifier=None
    ):
        super().__init__(
            rows, ultimate_receiver_id=ultimate_receiver_id,
            originator_id=originator_id, group_status=group_status,
            as_of_date=as_of_date, as_of_time=as_of_time, currency=currency,
            as_of_date_modifier=as_of_date_modifier
        )


class GroupTrailer(Bai2SingleModel):
    code = RecordCode.group_trailer

    def __init__(
        self,
        rows=None,
        group_control_total=None,
        number_of_accounts=None,
        number_of_records=None
    ):
        super().__init__(
            rows, group_control_total=group_control_total,
            number_of_accounts=number_of_accounts,
            number_of_records=number_of_records
        )


class Account(Bai2SectionModel):
    def __init__(self, header=None, trailer=None, children=None):
        super().__init__(header=header or AccountIdentifier([]),
                         trailer=trailer or AccountTrailer([]),
                         children=children or [])

    def update_totals(self):
        account_control_total = 0
        for transaction in self.children:
            account_control_total += transaction.amount
        for summary in self.header.summary_items:
            account_control_total += summary.amount

        self.trailer.account_control_total = account_control_total


class AccountIdentifier(Bai2SingleModel):
    code = RecordCode.account_identifier

    def __init__(
        self,
        rows=None,
        customer_account_number=None,
        currency=None,
        summary_items=()
    ):
        summary_items = list(summary_items)
        super().__init__(
            rows, customer_account_number=customer_account_number,
            currency=currency, summary_items=summary_items
        )


class Summary:
    def __init__(
        self,
        type_code=None,
        amount=0,
        item_count=None,
        funds_type=None,
        availability={}  # noqa: B006
    ):
        self.type_code = type_code
        self.amount = amount
        self.item_count = item_count
        self.funds_type = funds_type
        self.availability = availability


class AccountTrailer(Bai2SingleModel):
    code = RecordCode.account_trailer

    def __init__(
        self,
        rows=None,
        account_control_total=None,
        number_of_records=None
    ):
        super().__init__(
            rows, account_control_total=account_control_total,
            number_of_records=number_of_records
        )


class TransactionDetail(Bai2SingleModel):
    code = RecordCode.transaction_detail

    def __init__(
        self,
        rows=None,
        type_code=None,
        amount=None,
        funds_type=None,
        availability={},  # noqa: B006
        bank_reference=None,
        customer_reference=None,
        text=None
    ):
        super().__init__(
            rows, type_code=type_code, amount=amount, funds_type=funds_type,
            availability=availability, bank_reference=bank_reference,
            customer_reference=customer_reference, text=text
        )



import pathlib
from unittest import TestCase

from bai2 import bai2
from bai2.models import Bai2File

from tests.test_writers import Bai2FileWriterTestCase


class ParseTestCase(TestCase):
    @classmethod
    def open_test_file(cls, name):
        path = pathlib.Path(__file__).parent / 'data' / f'{name}.bai2'
        return path.open()

    def test_parse_from_lines(self):
        lines = [
            '01,CITIDIRECT,8888888,150716,0713,00131100,,,2/',
            '02,8888888,CITIGB00,1,150715,2340,GBP,2/',
            '03,77777777,GBP,010,10000,,,015,10000,,,/',
            '16,191,001,V,150715,,1234567890,RP12312312312312/',
            '88,FR:FP SIP INCOMING',
            '88,ENDT:20150715',
            '88,TRID:RP12312312312312',
            '88,PY:RP1231231231231200                 A1234BC 22/03/66',
            '88,BI:22222222',
            '88,OB:111111 BUCKINGHAM PALACE OB3:BARCLAYS BANK PLC',
            '88,BO:11111111 BO1:DOE JO',
            '49,20001,10/',
            '98,20001,1,12/',
            '99,20001,1,14/',
        ]

        bai2_file = bai2.parse_from_lines(lines)
        self.assertTrue(isinstance(bai2_file, Bai2File))

    def test_parse_from_string(self):
        s = (
            '01,CITIDIRECT,8888888,150716,0713,00131100,,,2/\n'
            '02,8888888,CITIGB00,1,150715,2340,GBP,2/\n'
            '03,77777777,GBP,010,10000,,,015,10000,,,/\n'
            '16,191,001,V,150715,,1234567890,RP12312312312312/\n'
            '88,FR:FP SIP INCOMING\n'
            '88,ENDT:20150715\n'
            '88,TRID:RP12312312312312\n'
            '88,PY:RP1231231231231200                 A1234BC 22/03/66\n'
            '88,BI:22222222\n'
            '88,OB:111111 BUCKINGHAM PALACE OB3:BARCLAYS BANK PLC\n'
            '88,BO:11111111 BO1:DOE JO\n'
            '49,20001,10/\n'
            '98,20001,1,12/\n'
            '99,20001,1,14/\n'
        )

        bai2_file = bai2.parse_from_string(s)
        self.assertTrue(isinstance(bai2_file, Bai2File))

    def test_parse_from_file(self):
        with self.open_test_file('citi_example') as f:
            bai2_file = bai2.parse_from_file(f)
            self.assertTrue(isinstance(bai2_file, Bai2File))

    def test_parse_from_file_2(self):
        with self.open_test_file('nwb_example') as f:
            bai2_file = bai2.parse_from_file(f)
            self.assertTrue(isinstance(bai2_file, Bai2File))

    def test_parse_from_file_3(self):
        with self.open_test_file('svb_us_example') as f:
            bai2_file = bai2.parse_from_file(f)
            self.assertTrue(isinstance(bai2_file, Bai2File))

    def test_parse_from_file_with_known_parsing_issue(self):
        with self.open_test_file('account_trailer_amount_blank_example') as f:
            bai2_file = bai2.parse_from_file(f)
            self.assertTrue(isinstance(bai2_file, Bai2File))

    def test_as_string(self):
        original = (
            '01,CITIDIRECT,8888888,150716,0713,00131100,,,2/\n'
            '02,8888888,CITIGB00,1,150715,2340,GBP,2/\n'
            '03,77777777,GBP,010,10000,,,015,10000,,,/\n'
            '16,191,001,V,150715,,1234567890,RP12312312312312/\n'
            '88,FR:FP SIP INCOMING\n'
            '88,ENDT:20150715\n'
            '88,TRID:RP12312312312312\n'
            '88,PY:RP1231231231231200                 A1234BC 22/03/66\n'
            '88,BI:22222222\n'
            '88,OB:111111 BUCKINGHAM PALACE OB3:BARCLAYS BANK PLC\n'
            '88,BO:11111111 BO1:DOE JO\n'
            '49,20001,10/\n'
            '98,20001,1,12/\n'
            '99,20001,1,14/'
        )

        bai2_file = bai2.parse_from_string(original)
        self.assertTrue(isinstance(bai2_file, Bai2File))

        from_model = bai2_file.as_string()

        self.assertEqual(original, from_model)


class WriteTestCase(TestCase):
    def test_write(self):
        bai2_file = Bai2FileWriterTestCase.create_bai2_file()

        output = bai2.write(bai2_file)
        self.assertEqual(
            output,
            (
                '01,CITIDIRECT,8888888,150715,2340,00131100,,,2/\n'
                '02,8888888,CITIGB00,1,150715,2340,GBP,2/\n'
                '03,77777777,GBP,010,10000,,,015,10000,,/\n'
                '16,399,2599,,,,BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS\n'
                '88, BILLS\n'
                '16,399,1000,0,,,OTHER\n'
                '49,23599,5/\n'
                '03,77777777,GBP,010,10000,,,015,10000,,/\n'
                '16,399,2599,,,,BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS\n'
                '88, BILLS\n'
                '16,399,1000,0,,,OTHER\n'
                '49,23599,5/\n'
                '98,47198,2,12/\n'
                '02,8888888,CITIGB00,1,150715,2340,GBP,2/\n'
                '03,77777777,GBP,010,10000,,,015,10000,,/\n'
                '16,399,2599,,,,BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS\n'
                '88, BILLS\n'
                '16,399,1000,0,,,OTHER\n'
                '49,23599,5/\n'
                '03,77777777,GBP,010,10000,,,015,10000,,/\n'
                '16,399,2599,,,,BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS\n'
                '88, BILLS\n'
                '16,399,1000,0,,,OTHER\n'
                '49,23599,5/\n'
                '98,47198,2,12/\n'
                '99,94396,2,26/'
            )
        )



from collections import OrderedDict

from .constants import GroupStatus, AsOfDateModifier, FundsType
from .exceptions import ParsingException, NotSupportedYetException, \
    IntegrityException
from .models import \
    Bai2File, Bai2FileHeader, Bai2FileTrailer, \
    Group, GroupHeader, GroupTrailer, \
    AccountIdentifier, AccountTrailer, Account, \
    TransactionDetail, Summary
from .utils import parse_date, parse_time, parse_type_code


# ABSTRACTION

class BaseParser:
    model = None
    child_parser_class = None

    def __init__(self, iterator, check_integrity=True):
        """
        Keyword arguments:
        check_integrity -- checks the data integrity of the parsed file (default True)
        """
        super().__init__()
        self._iter = iterator
        self.check_integrity = check_integrity

        self.child_parser = self._get_parser('child')

    def _check_record_code(self, expected_code):
        if self._iter.current_record.code != expected_code:
            raise ParsingException(
                'Expected {expected}, got {found} instead'.format(
                    expected=expected_code,
                    found=self._iter.current_record.code
                )
            )

    def _get_parser(self, parser_type):
        name = '{name}_parser_class'.format(name=parser_type.lower())
        parser_clazz = getattr(self, name)
        if parser_clazz:
            return parser_clazz(
                self._iter,
                check_integrity=self.check_integrity,
            )
        return None

    def validate(self, obj):
        pass

    def can_parse(self):
        raise NotImplementedError()

    def parse(self):
        raise NotImplementedError()


class BaseSectionParser(BaseParser):
    header_parser_class = None
    trailer_parser_class = None

    def __init__(self, iterator, **kwargs):
        super().__init__(iterator, **kwargs)

        self.header_parser = self._get_parser('header')
        self.trailer_parser = self._get_parser('trailer')

    def _parse_header(self):
        return self.header_parser.parse()

    def _parse_trailer(self):
        return self.trailer_parser.parse()

    def _parse_children(self):
        if not self.child_parser:
            return []

        children = []
        while self.child_parser.can_parse():
            children.append(self.child_parser.parse())

        return children

    def can_parse(self):
        return \
            self.header_parser.can_parse() or \
            self.trailer_parser.can_parse() or \
            (self.child_parser and self.child_parser.can_parse())

    def validate_number_of_records(self, obj):
        if self.check_integrity:
            number_of_records = len(obj.header.rows) + len(obj.trailer.rows)
            for child in obj.children:
                number_of_records += len(child.rows)

            if number_of_records != obj.trailer.number_of_records:
                raise IntegrityException(
                    'Invalid number of records for {clazz}. '
                    'expected {expected}, found {found}'.format(
                        clazz=obj.__class__.__name__,
                        expected=obj.trailer.number_of_records,
                        found=number_of_records
                    )
                )

    def validate(self, obj):
        super().validate(obj)

        self.validate_number_of_records(obj)

    def parse(self):
        header = self._parse_header()
        children = self._parse_children()
        trailer = self._parse_trailer()

        obj = self.build_model(header, children, trailer)

        self.validate(obj)

        return obj

    def build_model(self, header, children, trailer):
        return self.model(header, trailer, children)


class BaseSingleParser(BaseParser):
    fields_config = {}

    def can_parse(self):
        try:
            self._check_record_code(self.model.code)
        except ParsingException:
            return False
        return True

    def _parse_field_from_config(self, field_config, raw_value):
        if isinstance(field_config, str):
            field_config = (field_config, lambda x: x)

        field_name, parser = field_config

        # Integer check
        if parser == int and 'total' in field_name:
            field_value = parser(raw_value) if raw_value else 0
        else:
            field_value = parser(raw_value) if raw_value else None
        return field_name, field_value

    def _parse_fields_from_config(self, values, fields_config):
        fields = {}

        index = 0
        for field_config in fields_config:
            field_name, field_value = self._parse_field_from_config(
                field_config, values[index]
            )
            fields[field_name] = field_value
            index += 1
        return fields

    def _parse_fields(self, record):
        return self._parse_fields_from_config(
            record.fields, self.fields_config
        )

    def _parse_availability(self, funds_type, rest):
        availability = None
        if funds_type == FundsType.distributed_availability_simple:
            availability = OrderedDict()
            for day in ['0', '1', '>1']:
                availability[day] = int(rest.pop(0))
        elif funds_type == FundsType.value_dated:
            date = rest.pop(0)
            time = rest.pop(0)
            availability = OrderedDict()
            availability['date'] = parse_date(date) if date else None
            availability['time'] = parse_time(time) if time else None
        elif funds_type == FundsType.distributed_availability:
            num_distributions = int(rest.pop(0))
            availability = OrderedDict()
            for _ in range(num_distributions):
                day = rest.pop(0)
                amount = int(rest.pop(0))
                availability[day] = amount

        return availability, rest

    def parse(self):
        self._check_record_code(self.model.code)
        obj = self.model(
            self._iter.current_record.rows,
            **self._parse_fields(self._iter.current_record)
        )

        self.validate(obj)

        try:
            self._iter.advance()
        except StopIteration:
            pass

        return obj


# IMPLEMENTATION

class TransactionDetailParser(BaseSingleParser):
    model = TransactionDetail

    head_fields_config = [
        ('type_code', parse_type_code),
        ('amount', int),
        ('funds_type', FundsType),
    ]

    tail_fields_config = [
        'bank_reference',
        'customer_reference',
        'text',
    ]

    def _parse_fields(self, record):
        # fields at the start
        rest = record.fields
        fields = self._parse_fields_from_config(
            rest[:len(self.head_fields_config)],
            self.head_fields_config
        )

        rest = rest[len(self.head_fields_config):]
        # availability fields:
        availability, rest = self._parse_availability(
            fields['funds_type'], rest
        )
        fields['availability'] = availability

        # fields at the end
        fields.update(
            self._parse_fields_from_config(
                rest[:2] + [','.join(rest[2:])],
                self.tail_fields_config
            )
        )

        return fields


class AccountIdentifierParser(BaseSingleParser):
    model = AccountIdentifier

    common_fields_config = [
        'customer_account_number',
        'currency',
    ]
    summary_fields_config = [
        ('type_code', parse_type_code),
        ('amount', int),
        ('item_count', int),
        ('funds_type', FundsType),
    ]

    def _parse_fields(self, record):
        model_fields = self._parse_fields_from_config(
            record.fields[:len(self.common_fields_config)],
            self.common_fields_config
        )

        summary_items = []
        rest = record.fields[len(self.common_fields_config):]
        while rest:
            # there's currently a bug in some exports so we need to ignore
            # the last empty item if it's the only one left.
            if len(rest) == 1 and not rest[0]:
                break

            summary = self._parse_fields_from_config(
                rest, self.summary_fields_config
            )
            rest = rest[len(self.summary_fields_config):]
            availability, rest = self._parse_availability(
                summary['funds_type'], rest
            )
            if availability:
                summary['availability'] = availability
            summary_items.append(Summary(**summary))
        model_fields['summary_items'] = summary_items

        return model_fields


class AccountTrailerParser(BaseSingleParser):
    model = AccountTrailer

    fields_config = [
        ('account_control_total', int),
        ('number_of_records', int),
    ]


class AccountParser(BaseSectionParser):
    model = Account
    header_parser_class = AccountIdentifierParser
    trailer_parser_class = AccountTrailerParser
    child_parser_class = TransactionDetailParser

    def validate(self, obj):
        super().validate(obj)

        if self.check_integrity:
            transaction_sum = sum([child.amount or 0 for child in obj.children])
            account_sum = sum([summary.amount or 0 for summary in obj.header.summary_items])

            control_total = transaction_sum + account_sum
            if control_total != obj.trailer.account_control_total:
                raise IntegrityException(
                    'Invalid account control total for {clazz}. '
                    'expected {expected}, found {found}'.format(
                        clazz=obj.__class__.__name__,
                        expected=obj.trailer.account_control_total,
                        found=control_total
                    )
                )


class GroupHeaderParser(BaseSingleParser):
    model = GroupHeader

    fields_config = [
        'ultimate_receiver_id',
        'originator_id',
        ('group_status', GroupStatus),
        ('as_of_date', parse_date),
        ('as_of_time', parse_time),
        'currency',
        ('as_of_date_modifier', AsOfDateModifier),
    ]

    def _parse_fields(self, record):
        fields = super()._parse_fields(record)

        # if currency not defined => default to USD
        if not fields['currency']:
            fields['currency'] = 'USD'

        return fields


class GroupTrailerParser(BaseSingleParser):
    model = GroupTrailer

    fields_config = [
        ('group_control_total', int),
        ('number_of_accounts', int),
        ('number_of_records', int),
    ]


class GroupParser(BaseSectionParser):
    model = Group
    header_parser_class = GroupHeaderParser
    trailer_parser_class = GroupTrailerParser
    child_parser_class = AccountParser

    def validate(self, obj):
        super().validate(obj)

        if not obj.children:
            raise ParsingException('Group without accounts not allowed')

        if self.check_integrity:
            if obj.trailer.number_of_accounts != len(obj.children):
                raise IntegrityException(
                    'Invalid number of accounts for {clazz}. '
                    'expected {expected}, found {found}'.format(
                        clazz=obj.__class__.__name__,
                        expected=obj.trailer.number_of_accounts,
                        found=len(obj.children)
                    )
                )

            control_total = sum([
                account.trailer.account_control_total
                for account in obj.children
            ])

            if control_total != obj.trailer.group_control_total:
                raise IntegrityException(
                    'Invalid group control total for {clazz}. '
                    'expected {expected}, found {found}'.format(
                        clazz=obj.__class__.__name__,
                        expected=obj.trailer.group_control_total,
                        found=control_total
                    )
                )


class Bai2FileHeaderParser(BaseSingleParser):
    model = Bai2FileHeader

    fields_config = (
        'sender_id',
        'receiver_id',
        ('creation_date', parse_date),
        ('creation_time', parse_time),
        'file_id',
        ('physical_record_length', int),
        ('block_size', int),
        ('version_number', int),
    )

    def validate(self, obj):
        super().validate(obj)

        if obj.version_number != 2:
            raise NotSupportedYetException(
                'Only BAI version 2 supported'
            )


class Bai2FileTrailerParser(BaseSingleParser):
    model = Bai2FileTrailer

    fields_config = (
        ('file_control_total', int),
        ('number_of_groups', int),
        ('number_of_records', int),
    )


class Bai2FileParser(BaseSectionParser):
    model = Bai2File
    header_parser_class = Bai2FileHeaderParser
    trailer_parser_class = Bai2FileTrailerParser
    child_parser_class = GroupParser

    def validate(self, obj):
        super().validate(obj)

        if not obj.children:
            raise ParsingException('File without groups not allowed')

        if self.check_integrity:
            if obj.trailer.number_of_groups != len(obj.children):
                raise IntegrityException(
                    'Invalid number of groups for {clazz}. '
                    'expected {expected}, found {found}'.format(
                        clazz=obj.__class__.__name__,
                        expected=obj.trailer.number_of_groups,
                        found=len(obj.children)
                    )
                )

            control_total = sum([
                account.trailer.group_control_total
                for account in obj.children
            ])

            if control_total != obj.trailer.file_control_total:
                raise IntegrityException(
                    'Invalid file control total for {clazz}. '
                    'expected {expected}, found {found}'.format(
                        clazz=obj.__class__.__name__,
                        expected=obj.trailer.file_control_total,
                        found=control_total
                    )
                )



import datetime
from unittest import TestCase

from bai2.utils import parse_date, parse_military_time, parse_time, write_time


class ParseDateTestCase(TestCase):
    def test_parse(self):
        parsed = parse_date('150330')
        self.assertEqual(
            datetime.date(year=2015, month=3, day=30),
            parsed
        )


class ParseTimeTestCase(TestCase):
    def test_clock_time(self):
        parsed_value = parse_time('21:45:34')

        self.assertEqual(
            datetime.time(hour=21, minute=45, second=34),
            parsed_value
        )

    def test_military_time(self):
        parsed_value = parse_time('2145')

        self.assertEqual(
            datetime.time(hour=21, minute=45),
            parsed_value
        )


class ParseMilitaryTime(TestCase):
    def test_parse(self):
        parsed_value = parse_military_time('2145')

        self.assertEqual(
            datetime.time(hour=21, minute=45),
            parsed_value
        )

    def test_parse_end_of_day_as_9999(self):
        parsed_value = parse_military_time('9999')

        self.assertEqual(
            datetime.time.max,
            parsed_value
        )

    def test_parse_end_of_day_as_2400(self):
        parsed_value = parse_military_time('2400')

        self.assertEqual(
            datetime.time.max,
            parsed_value
        )

    def test_parse_beginning_of_day(self):
        parsed_value = parse_military_time('0000')

        self.assertEqual(
            datetime.time(hour=0, minute=0),
            parsed_value
        )


class WriteTime(TestCase):
    def test_write_intra_day_time_with_setting_false(self):
        time = datetime.time(hour=17, minute=59)

        str_value = write_time(time, False)
        self.assertEqual(str_value, '1759')

    def test_write_intra_day_time_with_setting_true(self):
        time = datetime.time(hour=17, minute=59, second=0)

        str_value = write_time(time, True)
        self.assertEqual(str_value, '17:59:00')

    def test_write_end_of_day_time_with_setting_false(self):
        time = datetime.time.max

        str_value = write_time(time, False)
        self.assertEqual(str_value, '2400')

    def test_write_end_of_day_time_with_setting_true(self):
        time = datetime.time.max

        str_value = write_time(time, True)
        self.assertEqual(str_value, '2400')



from bai2.parsers import Bai2FileParser
from bai2.writers import Bai2FileWriter
from bai2.helpers import IteratorHelper


def parse_from_lines(lines, **kwargs):
    helper = IteratorHelper(lines)
    parser = Bai2FileParser(helper, **kwargs)
    return parser.parse()


def parse_from_string(s, **kwargs):
    return parse_from_lines(s.splitlines(), **kwargs)


def parse_from_file(f, **kwargs):
    return parse_from_string(f.read(), **kwargs)


def write(bai2_obj, **kwargs):
    return '\n'.join(Bai2FileWriter(bai2_obj, **kwargs).write())



from collections import namedtuple
from enum import Enum

CONTINUATION_CODE = '88'


class RecordCode(Enum):
    file_header = '01'
    group_header = '02'
    account_identifier = '03'
    transaction_detail = '16'
    account_trailer = '49'
    continuation = '88'
    group_trailer = '98'
    file_trailer = '99'


class GroupStatus(Enum):
    update = '1'
    deletion = '2'
    correction = '3'
    test_only = '4'


class AsOfDateModifier(Enum):
    interim_previous_day = '1'
    final_previous_day = '2'
    interim_same_day = '3'
    final_same_day = '4'


class FundsType(Enum):
    unknown_availability = 'Z'
    immediate_availability = '0'
    one_day_availability = '1'
    two_day_availability = '2'
    distributed_availability_simple = 'S'
    value_dated = 'V'
    distributed_availability = 'D'


class TypeCodeTransaction(Enum):
    credit = 'credit'
    debit = 'debit'
    misc = 'misc'


class TypeCodeLevel(Enum):
    status = 'status'
    detail = 'detail'
    summary = 'summary'


TypeCode = namedtuple('TypeCode', ['code', 'transaction', 'level', 'description'])
TypeCodes = [
    TypeCode('010', None, TypeCodeLevel.status, 'Opening Ledger'),
    TypeCode('011', None, TypeCodeLevel.status, 'Average Opening Ledger MTD'),
    TypeCode('012', None, TypeCodeLevel.status, 'Average Opening Ledger YTD'),
    TypeCode('015', None, TypeCodeLevel.status, 'Closing Ledger'),
    TypeCode('020', None, TypeCodeLevel.status, 'Average Closing Ledger MTD'),
    TypeCode('021', None, TypeCodeLevel.status, 'Average Closing Ledger – Previous Month'),
    TypeCode('022', None, TypeCodeLevel.status, 'Aggregate Balance Adjustments'),
    TypeCode('024', None, TypeCodeLevel.status, 'Average Closing Ledger YTD – Previous Month'),
    TypeCode('025', None, TypeCodeLevel.status, 'Average Closing Ledger YTD'),
    TypeCode('030', None, TypeCodeLevel.status, 'Current Ledger'),
    TypeCode('037', None, TypeCodeLevel.status, 'ACH Net Position'),
    TypeCode('039', None, TypeCodeLevel.status, 'Opening Available + Total Same-Day ACH DTC Deposit'),
    TypeCode('040', None, TypeCodeLevel.status, 'Opening Available'),
    TypeCode('041', None, TypeCodeLevel.status, 'Average Opening Available MTD'),
    TypeCode('042', None, TypeCodeLevel.status, 'Average Opening Available YTD'),
    TypeCode('043', None, TypeCodeLevel.status, 'Average Available – Previous Month'),
    TypeCode('044', None, TypeCodeLevel.status, 'Disbursing Opening Available Balance'),
    TypeCode('045', None, TypeCodeLevel.status, 'Closing Available'),
    TypeCode('050', None, TypeCodeLevel.status, 'Average Closing Available MTD'),
    TypeCode('051', None, TypeCodeLevel.status, 'Average Closing Available – Last Month'),
    TypeCode('054', None, TypeCodeLevel.status, 'Average Closing Available YTD – Last Month'),
    TypeCode('055', None, TypeCodeLevel.status, 'Average Closing Available YTD'),
    TypeCode('056', None, TypeCodeLevel.status, 'Loan Balance'),
    TypeCode('057', None, TypeCodeLevel.status, 'Total Investment Position'),
    TypeCode('059', None, TypeCodeLevel.status, 'Current Available (CRS Supressed)'),
    TypeCode('060', None, TypeCodeLevel.status, 'Current Available'),
    TypeCode('061', None, TypeCodeLevel.status, 'Average Current Available MTD'),
    TypeCode('062', None, TypeCodeLevel.status, 'Average Current Available YTD'),
    TypeCode('063', None, TypeCodeLevel.status, 'Total Float'),
    TypeCode('065', None, TypeCodeLevel.status, 'Target Balance'),
    TypeCode('066', None, TypeCodeLevel.status, 'Adjusted Balance'),
    TypeCode('067', None, TypeCodeLevel.status, 'Adjusted Balance MTD'),
    TypeCode('068', None, TypeCodeLevel.status, 'Adjusted Balance YTD'),
    TypeCode('070', None, TypeCodeLevel.status, '0-Day Float'),
    TypeCode('072', None, TypeCodeLevel.status, '1-Day Float'),
    TypeCode('073', None, TypeCodeLevel.status, 'Float Adjustment'),
    TypeCode('074', None, TypeCodeLevel.status, '2 or More Days Float'),
    TypeCode('075', None, TypeCodeLevel.status, '3 or More Days Float'),
    TypeCode('076', None, TypeCodeLevel.status, 'Adjustment to Balances'),
    TypeCode('077', None, TypeCodeLevel.status, 'Average Adjustment to Balances MTD'),
    TypeCode('078', None, TypeCodeLevel.status, 'Average Adjustment to Balances YTD'),
    TypeCode('079', None, TypeCodeLevel.status, '4-Day Float'),
    TypeCode('080', None, TypeCodeLevel.status, '5-Day Float'),
    TypeCode('081', None, TypeCodeLevel.status, '6-Day Float'),
    TypeCode('082', None, TypeCodeLevel.status, 'Average 1-Day Float MTD'),
    TypeCode('083', None, TypeCodeLevel.status, 'Average 1-Day Float YTD'),
    TypeCode('084', None, TypeCodeLevel.status, 'Average 2-Day Float MTD'),
    TypeCode('085', None, TypeCodeLevel.status, 'Average 2-Day Float YTD'),
    TypeCode('086', None, TypeCodeLevel.status, 'Transfer Calculation'),
    TypeCode('087', None, TypeCodeLevel.status, 'MTD Ave Collected'),
    TypeCode('088', None, TypeCodeLevel.status, 'MTD Ave Neg Collected'),
    TypeCode('100', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Credits'),
    TypeCode('101', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Credit Amount MTD'),
    TypeCode('102', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Number of Credits'),
    TypeCode('105', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Credits Not Detailed'),
    TypeCode('106', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Deposits Subject to Float'),
    TypeCode('107', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Adjustment Credits YTD'),
    TypeCode('108', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Credit (Any Type)'),
    TypeCode('109', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Current Day Total Lockbox Deposits'),
    TypeCode('110', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Lockbox Deposits'),
    TypeCode('115', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Lockbox Deposit'),
    TypeCode('116', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Item in Lockbox Deposit'),
    TypeCode('118', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Lockbox Adjustment Credit'),
    TypeCode('120', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'EDI* Transaction Credit'),
    TypeCode('121', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'EDI Transaction Credit'),
    TypeCode('122', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'EDIBANX Credit Received'),
    TypeCode('123', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'EDIBANX Credit Return'),
    TypeCode('130', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Concentration Credits'),
    TypeCode('131', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total DTC Credits'),
    TypeCode('135', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'DTC Concentration Credit'),
    TypeCode('136', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Item in DTC Deposit'),
    TypeCode('140', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total ACH Credits'),
    TypeCode('142', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'ACH Credit Received'),
    TypeCode('143', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Item in ACH Deposit'),
    TypeCode('145', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'ACH Concentration Credit'),
    TypeCode('146', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Bank Card Deposits'),
    TypeCode('147', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Individual Bank Card Deposit'),
    TypeCode('150', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Preauthorized Payment Credits'),
    TypeCode('155', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Preauthorized Draft Credit'),
    TypeCode('156', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Item in PAC Deposit'),
    TypeCode('158', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Real Time Payment Credit'),
    TypeCode('159', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Real Time Payment Credits'),
    TypeCode('160', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total ACH Disbursing Funding Credits'),
    TypeCode('162', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Corporate Trade Payment Settlement'),
    TypeCode('163', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Corporate Trade Payment Credits'),
    TypeCode('164', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Corporate Trade Payment Credit'),
    TypeCode('165', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Preauthorized ACH Credit'),
    TypeCode('166', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'ACH Settlement'),
    TypeCode('167', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'ACH Settlement Credits'),
    TypeCode('168', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'ACH Return Item or Adjustment Settlement'),
    TypeCode('169', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Miscellaneous ACH Credit'),
    TypeCode('170', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Other Check Deposits'),
    TypeCode('171', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Individual Loan Deposit'),
    TypeCode('172', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Deposit Correction'),
    TypeCode('173', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Bank-Prepared Deposit'),
    TypeCode('174', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Other Deposit'),
    TypeCode('175', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Check Deposit Package'),
    TypeCode('176', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Re-presented Check Deposit'),
    TypeCode('178', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'List Post Credits'),
    TypeCode('180', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Loan Proceeds'),
    TypeCode('182', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Bank-Prepared Deposits'),
    TypeCode('184', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Draft Deposit'),
    TypeCode('185', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Miscellaneous Deposits'),
    TypeCode('186', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Cash Letter Credits'),
    TypeCode('187', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Cash Letter Credit'),
    TypeCode('188', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Cash Letter Adjustments'),
    TypeCode('189', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Cash Letter Adjustment'),
    TypeCode('190', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Incoming Money Transfers'),
    TypeCode('191', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Individual Incoming Internal Money Transfer'),
    TypeCode('195', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Incoming Money Transfer'),
    TypeCode('196', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Money Transfer Adjustment'),
    TypeCode('198', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Compensation'),
    TypeCode('200', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Automatic Transfer Credits'),
    TypeCode('201', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Individual Automatic Transfer Credit'),
    TypeCode('202', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Bond Operations Credit'),
    TypeCode('205', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Book Transfer Credits'),
    TypeCode('206', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Book Transfer Credit'),
    TypeCode('207', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total International Money Transfer Credits'),
    TypeCode('208', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Individual International Money Transfer Credit'),
    TypeCode('210', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total International Credits'),
    TypeCode('212', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Foreign Letter of Credit'),
    TypeCode('213', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Letter of Credit'),
    TypeCode('214', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Foreign Exchange of Credit'),
    TypeCode('215', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Letters of Credit'),
    TypeCode('216', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Foreign Remittance Credit'),
    TypeCode('218', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Foreign Collection Credit'),
    TypeCode('221', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Foreign Check Purchase'),
    TypeCode('222', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Foreign Checks Deposited'),
    TypeCode('224', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Commission'),
    TypeCode('226', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'International Money Market Trading'),
    TypeCode('227', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Standing Order'),
    TypeCode('229', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Miscellaneous International Credit'),
    TypeCode('230', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Security Credits'),
    TypeCode('231', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Collection Credits'),
    TypeCode('232', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Sale of Debt Security'),
    TypeCode('233', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Securities Sold'),
    TypeCode('234', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Sale of Equity Security'),
    TypeCode('235', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Matured Reverse Repurchase Order'),
    TypeCode('236', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Maturity of Debt Security'),
    TypeCode('237', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Individual Collection Credit'),
    TypeCode('238', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Collection of Dividends'),
    TypeCode('239', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Bankers’ Acceptance Credits'),
    TypeCode('240', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Coupon Collections – Banks'),
    TypeCode('241', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Bankers’ Acceptances'),
    TypeCode('242', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Collection of Interest Income'),
    TypeCode('243', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Matured Fed Funds Purchased'),
    TypeCode('244', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Interest/Matured Principal Payment'),
    TypeCode('245', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Monthly Dividends'),
    TypeCode('246', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Commercial Paper'),
    TypeCode('247', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Capital Change'),
    TypeCode('248', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Savings Bonds Sales Adjustment'),
    TypeCode('249', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Miscellaneous Security Credit'),
    TypeCode('250', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Checks Posted and Returned'),
    TypeCode('251', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Debit Reversals'),
    TypeCode('252', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Debit Reversal'),
    TypeCode('254', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Posting Error Correction Credit'),
    TypeCode('255', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Check Posted and Returned'),
    TypeCode('256', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total ACH Return Items'),
    TypeCode('257', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Individual ACH Return Item'),
    TypeCode('258', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'ACH Reversal Credit'),
    TypeCode('260', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Rejected Credits'),
    TypeCode('261', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Individual Rejected Credit'),
    TypeCode('263', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Overdraft'),
    TypeCode('266', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Return Item'),
    TypeCode('268', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Return Item Adjustment'),
    TypeCode('270', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total ZBA Credits'),
    TypeCode('271', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Net Zero-Balance Amount'),
    TypeCode('274', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Cumulative** ZBA or Disbursement Credits'),
    TypeCode('275', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'ZBA Credit'),
    TypeCode('276', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'ZBA Float Adjustment'),
    TypeCode('277', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'ZBA Credit Transfer'),
    TypeCode('278', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'ZBA Credit Adjustment'),
    TypeCode('280', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Controlled Disbursing Credits'),
    TypeCode('281', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Individual Controlled Disbursing Credit'),
    TypeCode('285', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total DTC Disbursing Credits'),
    TypeCode('286', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Individual DTC Disbursing Credit'),
    TypeCode('294', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total ATM Credits'),
    TypeCode('295', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'ATM Credit'),
    TypeCode('301', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Commercial Deposit'),
    TypeCode('302', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Correspondent Bank Deposit'),
    TypeCode('303', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Wire Transfers In – FF'),
    TypeCode('304', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Wire Transfers In – CHF'),
    TypeCode('305', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Fed Funds Sold'),
    TypeCode('306', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Fed Funds Sold'),
    TypeCode('307', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Trust Credits'),
    TypeCode('308', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Trust Credit'),
    TypeCode('309', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Value - Dated Funds'),
    TypeCode('310', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Commercial Deposits'),
    TypeCode('315', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total International Credits – FF'),
    TypeCode('316', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total International Credits – CHF'),
    TypeCode('318', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Foreign Check Purchased'),
    TypeCode('319', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Late Deposit'),
    TypeCode('320', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Securities Sold – FF'),
    TypeCode('321', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Securities Sold – CHF'),
    TypeCode('324', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Securities Matured – FF'),
    TypeCode('325', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Securities Matured – CHF'),
    TypeCode('326', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Securities Interest'),
    TypeCode('327', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Securities Matured'),
    TypeCode('328', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Securities Interest – FF'),
    TypeCode('329', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Securities Interest – CHF'),
    TypeCode('330', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Escrow Credits'),
    TypeCode('331', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Individual Escrow Credit'),
    TypeCode('332', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Miscellaneous Securities Credits – FF'),
    TypeCode('336', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Miscellaneous Securities Credits – CHF'),
    TypeCode('338', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Securities Sold'),
    TypeCode('340', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Broker Deposits'),
    TypeCode('341', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Broker Deposits – FF'),
    TypeCode('342', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Broker Deposit'),
    TypeCode('343', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Broker Deposits – CHF'),
    TypeCode('344', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Individual Back Value Credit'),
    TypeCode('345', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Item in Brokers Deposit'),
    TypeCode('346', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Sweep Interest Income'),
    TypeCode('347', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Sweep Principal Sell'),
    TypeCode('348', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Futures Credit'),
    TypeCode('349', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Principal Payments Credit'),
    TypeCode('350', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Investment Sold'),
    TypeCode('351', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Individual Investment Sold'),
    TypeCode('352', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Cash Center Credits'),
    TypeCode('353', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Cash Center Credit'),
    TypeCode('354', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Interest Credit'),
    TypeCode('355', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Investment Interest'),
    TypeCode('356', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Credit Adjustment'),
    TypeCode('357', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Credit Adjustment'),
    TypeCode('358', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'YTD Adjustment Credit'),
    TypeCode('359', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Interest Adjustment Credit'),
    TypeCode('360', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Credits Less Wire Transfer and '
                                                                       'Returned Checks'),
    TypeCode('361', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Grand Total Credits Less Grand Total Debits'),
    TypeCode('362', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Correspondent Collection'),
    TypeCode('363', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Correspondent Collection Adjustment'),
    TypeCode('364', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Loan Participation'),
    TypeCode('366', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Currency and Coin Deposited'),
    TypeCode('367', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Food Stamp Letter'),
    TypeCode('368', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Food Stamp Adjustment'),
    TypeCode('369', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Clearing Settlement Credit'),
    TypeCode('370', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Back Value Credits'),
    TypeCode('372', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Back Value Adjustment'),
    TypeCode('373', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Customer Payroll'),
    TypeCode('374', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'FRB Statement Recap'),
    TypeCode('376', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Savings Bond Letter or Adjustment'),
    TypeCode('377', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Treasury Tax and Loan Credit'),
    TypeCode('378', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Transfer of Treasury Credit'),
    TypeCode('379', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'FRB Government Checks Cash Letter Credit'),
    TypeCode('381', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'FRB Government Check Adjustment'),
    TypeCode('382', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'FRB Postal Money Order Credit'),
    TypeCode('383', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'FRB Postal Money Order Adjustment'),
    TypeCode('384', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'FRB Cash Letter Auto Charge Credit'),
    TypeCode('385', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Universal Credits'),
    TypeCode('386', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'FRB Cash Letter Auto Charge Adjustment'),
    TypeCode('387', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'FRB Fine-Sort Cash Letter Credit'),
    TypeCode('388', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'FRB Fine-Sort Adjustment'),
    TypeCode('389', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Freight Payment Credits'),
    TypeCode('390', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Miscellaneous Credits'),
    TypeCode('391', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Universal Credit'),
    TypeCode('392', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Freight Payment Credit'),
    TypeCode('393', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Itemized Credit Over $10,000'),
    TypeCode('394', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Cumulative** Credits'),
    TypeCode('395', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Check Reversal'),
    TypeCode('397', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Float Adjustment'),
    TypeCode('398', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Miscellaneous Fee Refund'),
    TypeCode('399', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Miscellaneous Credit'),
    TypeCode('400', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Debits'),
    TypeCode('401', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Debit Amount MTD'),
    TypeCode('402', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Number of Debits'),
    TypeCode('403', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Today’s Total Debits'),
    TypeCode('405', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Debit Less Wire Transfers and '
                                                                      'ChargeBacks'),
    TypeCode('406', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Debits not Detailed'),
    TypeCode('408', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Float Adjustment'),
    TypeCode('409', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Debit (Any Type)'),
    TypeCode('410', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total YTD Adjustment'),
    TypeCode('412', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Debits (Excluding Returned Items)'),
    TypeCode('415', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Lockbox Debit'),
    TypeCode('416', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Lockbox Debits'),
    TypeCode('420', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'EDI Transaction Debits'),
    TypeCode('421', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'EDI Transaction Debit'),
    TypeCode('422', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'EDIBANX Settlement Debit'),
    TypeCode('423', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'EDIBANX Return Item Debit'),
    TypeCode('430', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Payable–Through Drafts'),
    TypeCode('435', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Payable–Through Draft'),
    TypeCode('445', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'ACH Concentration Debit'),
    TypeCode('446', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total ACH Disbursement Funding Debits'),
    TypeCode('447', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'ACH Disbursement Funding Debit'),
    TypeCode('450', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total ACH Debits'),
    TypeCode('451', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'ACH Debit Received'),
    TypeCode('452', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Item in ACH Disbursement or Debit'),
    TypeCode('455', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Preauthorized ACH Debit'),
    TypeCode('458', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Real Time Payment Debit'),
    TypeCode('459', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Real Time Payment Debits'),
    TypeCode('462', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Account Holder Initiated ACH Debit'),
    TypeCode('463', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Corporate Trade Payment Debits'),
    TypeCode('464', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Corporate Trade Payment Debit'),
    TypeCode('465', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Corporate Trade Payment Settlement'),
    TypeCode('466', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'ACH Settlement'),
    TypeCode('467', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'ACH Settlement Debits'),
    TypeCode('468', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'ACH Return Item or Adjustment Settlement'),
    TypeCode('469', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Miscellaneous ACH Debit'),
    TypeCode('470', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Check Paid'),
    TypeCode('471', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Check Paid – Cumulative MTD'),
    TypeCode('472', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Cumulative** Checks Paid'),
    TypeCode('474', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Certified Check Debit'),
    TypeCode('475', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Check Paid'),
    TypeCode('476', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Federal Reserve Bank Letter Debit'),
    TypeCode('477', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Bank Originated Debit'),
    TypeCode('478', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'List Post Debits'),
    TypeCode('479', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'List Post Debit'),
    TypeCode('480', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Loan Payments'),
    TypeCode('481', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Individual Loan Payment'),
    TypeCode('482', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Bank-Originated Debits'),
    TypeCode('484', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Draft'),
    TypeCode('485', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'DTC Debit'),
    TypeCode('486', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Cash Letter Debits'),
    TypeCode('487', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Cash Letter Debit'),
    TypeCode('489', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Cash Letter Adjustment'),
    TypeCode('490', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Outgoing Money Transfers'),
    TypeCode('491', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Individual Outgoing Internal Money Transfer'),
    TypeCode('493', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Customer Terminal Initiated Money Transfer'),
    TypeCode('495', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Outgoing Money Transfer'),
    TypeCode('496', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Money Transfer Adjustment'),
    TypeCode('498', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Compensation'),
    TypeCode('500', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Automatic Transfer Debits'),
    TypeCode('501', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Individual Automatic Transfer Debit'),
    TypeCode('502', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Bond Operations Debit'),
    TypeCode('505', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Book Transfer Debits'),
    TypeCode('506', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Book Transfer Debit'),
    TypeCode('507', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total International Money Transfer Debits'),
    TypeCode('508', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Individual International Money Transfer Debits'),
    TypeCode('510', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total International Debits'),
    TypeCode('512', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Letter of Credit Debit'),
    TypeCode('513', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Letter of Credit'),
    TypeCode('514', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Foreign Exchange Debit'),
    TypeCode('515', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Letters of Credit'),
    TypeCode('516', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Foreign Remittance Debit'),
    TypeCode('518', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Foreign Collection Debit'),
    TypeCode('522', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Foreign Checks Paid'),
    TypeCode('524', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Commission'),
    TypeCode('526', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'International Money Market Trading'),
    TypeCode('527', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Standing Order'),
    TypeCode('529', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Miscellaneous International Debit'),
    TypeCode('530', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Security Debits'),
    TypeCode('531', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Securities Purchased'),
    TypeCode('532', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Amount of Securities Purchased'),
    TypeCode('533', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Security Collection Debit'),
    TypeCode('534', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Miscellaneous Securities DB – FF'),
    TypeCode('535', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Purchase of Equity Securities'),
    TypeCode('536', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Miscellaneous Securities Debit – CHF'),
    TypeCode('537', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Collection Debit'),
    TypeCode('538', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Matured Repurchase Order'),
    TypeCode('539', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Bankers’ Acceptances Debit'),
    TypeCode('540', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Coupon Collection Debit'),
    TypeCode('541', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Bankers’ Acceptances'),
    TypeCode('542', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Purchase of Debt Securities'),
    TypeCode('543', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Domestic Collection'),
    TypeCode('544', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Interest/Matured Principal Payment'),
    TypeCode('546', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Commercial paper'),
    TypeCode('547', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Capital Change'),
    TypeCode('548', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Savings Bonds Sales Adjustment'),
    TypeCode('549', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Miscellaneous Security Debit'),
    TypeCode('550', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Deposited Items Returned'),
    TypeCode('551', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Credit Reversals'),
    TypeCode('552', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Credit Reversal'),
    TypeCode('554', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Posting Error Correction Debit'),
    TypeCode('555', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Deposited Item Returned'),
    TypeCode('556', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total ACH Return Items'),
    TypeCode('557', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Individual ACH Return Item'),
    TypeCode('558', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'ACH Reversal Debit'),
    TypeCode('560', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Rejected Debits'),
    TypeCode('561', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Individual Rejected Debit'),
    TypeCode('563', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Overdraft'),
    TypeCode('564', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Overdraft Fee'),
    TypeCode('566', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Return Item'),
    TypeCode('567', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Return Item Fee'),
    TypeCode('568', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Return Item Adjustment'),
    TypeCode('570', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total ZBA Debits'),
    TypeCode('574', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Cumulative ZBA Debits'),
    TypeCode('575', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'ZBA Debit'),
    TypeCode('577', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'ZBA Debit Transfer'),
    TypeCode('578', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'ZBA Debit Adjustment'),
    TypeCode('580', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Controlled Disbursing Debits'),
    TypeCode('581', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Individual Controlled Disbursing Debit'),
    TypeCode('583', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Disbursing Checks Paid – Early Amount'),
    TypeCode('584', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Disbursing Checks Paid – Later Amount'),
    TypeCode('585', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Disbursing Funding Requirement'),
    TypeCode('586', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'FRB Presentment Estimate (Fed Estimate)'),
    TypeCode('587', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Late Debits (After Notification)'),
    TypeCode('588', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Disbursing Checks Paid-Last Amount'),
    TypeCode('590', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total DTC Debits'),
    TypeCode('594', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total ATM Debits'),
    TypeCode('595', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'ATM Debit'),
    TypeCode('596', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total APR Debits'),
    TypeCode('597', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'ARP Debit'),
    TypeCode('601', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Estimated Total Disbursement'),
    TypeCode('602', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Adjusted Total Disbursement'),
    TypeCode('610', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Funds Required'),
    TypeCode('611', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Wire Transfers Out- CHF'),
    TypeCode('612', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Wire Transfers Out – FF'),
    TypeCode('613', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total International Debit – CHF'),
    TypeCode('614', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total International Debit – FF'),
    TypeCode('615', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Federal Reserve Bank – '
                                                                      'Commercial Bank Debit'),
    TypeCode('616', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Federal Reserve Bank – Commercial Bank Debit'),
    TypeCode('617', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Securities Purchased – CHF'),
    TypeCode('618', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Securities Purchased – FF'),
    TypeCode('621', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Broker Debits – CHF'),
    TypeCode('622', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Broker Debit'),
    TypeCode('623', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Broker Debits – FF'),
    TypeCode('625', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Broker Debits'),
    TypeCode('626', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Fed Funds Purchased'),
    TypeCode('627', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Fed Funds Purchased'),
    TypeCode('628', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Cash Center Debits'),
    TypeCode('629', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Cash Center Debit'),
    TypeCode('630', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Debit Adjustments'),
    TypeCode('631', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Debit Adjustment'),
    TypeCode('632', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Trust Debits'),
    TypeCode('633', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Trust Debit'),
    TypeCode('634', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'YTD Adjustment Debit'),
    TypeCode('640', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Escrow Debits'),
    TypeCode('641', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Individual Escrow Debit'),
    TypeCode('644', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Individual Back Value Debit'),
    TypeCode('646', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Transfer Calculation Debit'),
    TypeCode('650', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Investments Purchased'),
    TypeCode('651', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Individual Investment purchased'),
    TypeCode('654', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Interest Debit'),
    TypeCode('655', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Investment Interest Debits'),
    TypeCode('656', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Sweep Principal Buy'),
    TypeCode('657', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Futures Debit'),
    TypeCode('658', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Principal Payments Debit'),
    TypeCode('659', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Interest Adjustment Debit'),
    TypeCode('661', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Account Analysis Fee'),
    TypeCode('662', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Correspondent Collection Debit'),
    TypeCode('663', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Correspondent Collection Adjustment'),
    TypeCode('664', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Loan Participation'),
    TypeCode('665', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Intercept Debits'),
    TypeCode('666', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Currency and Coin Shipped'),
    TypeCode('667', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Food Stamp Letter'),
    TypeCode('668', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Food Stamp Adjustment'),
    TypeCode('669', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Clearing Settlement Debit'),
    TypeCode('670', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Back Value Debits'),
    TypeCode('672', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Back Value Adjustment'),
    TypeCode('673', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Customer Payroll'),
    TypeCode('674', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'FRB Statement Recap'),
    TypeCode('676', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Savings Bond Letter or Adjustment'),
    TypeCode('677', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Treasury Tax and Loan Debit'),
    TypeCode('678', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Transfer of Treasury Debit'),
    TypeCode('679', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'FRB Government Checks Cash Letter Debit'),
    TypeCode('681', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'FRB Government Check Adjustment'),
    TypeCode('682', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'FRB Postal Money Order Debit'),
    TypeCode('683', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'FRB Postal Money Order Adjustment'),
    TypeCode('684', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'FRB Cash Letter Auto Charge Debit'),
    TypeCode('685', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Universal Debits'),
    TypeCode('686', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'FRB Cash Letter Auto Charge Adjustment'),
    TypeCode('687', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'FRB Fine-Sort Cash Letter Debit'),
    TypeCode('688', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'FRB Fine-Sort Adjustment'),
    TypeCode('689', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'FRB Freight Payment Debits'),
    TypeCode('690', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Total Miscellaneous Debits'),
    TypeCode('691', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Universal Debit'),
    TypeCode('692', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Freight Payment Debit'),
    TypeCode('693', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Itemized Debit Over $10,000'),
    TypeCode('694', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Deposit Reversal'),
    TypeCode('695', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Deposit Correction Debit'),
    TypeCode('696', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Regular Collection Debit'),
    TypeCode('697', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Cumulative** Debits'),
    TypeCode('698', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Miscellaneous Fees'),
    TypeCode('699', TypeCodeTransaction.debit, TypeCodeLevel.detail, 'Miscellaneous Debit'),
    TypeCode('701', None, TypeCodeLevel.status, 'Principal Loan Balance'),
    TypeCode('703', None, TypeCodeLevel.status, 'Available Commitment Amount'),
    TypeCode('705', None, TypeCodeLevel.status, 'Payment Amount Due'),
    TypeCode('707', None, TypeCodeLevel.status, 'Principal Amount Past Due'),
    TypeCode('709', None, TypeCodeLevel.status, 'Interest Amount Past Due'),
    TypeCode('720', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Total Loan Payment'),
    TypeCode('721', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Amount Applied to Interest'),
    TypeCode('722', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Amount Applied to Principal'),
    TypeCode('723', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Amount Applied to Escrow'),
    TypeCode('724', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Amount Applied to Late Charges'),
    TypeCode('725', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Amount Applied to Buydown'),
    TypeCode('726', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Amount Applied to Misc. Fees'),
    TypeCode('727', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Amount Applied to Deferred Interest Detail'),
    TypeCode('728', TypeCodeTransaction.credit, TypeCodeLevel.detail, 'Amount Applied to Service Charge'),
    TypeCode('760', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Loan Disbursement'),
    TypeCode('827', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Incoming SEPA Payments'),
    TypeCode('828', TypeCodeTransaction.debit, TypeCodeLevel.summary, 'Outgoing SEPA Payments'),
    TypeCode('829', TypeCodeTransaction.credit, TypeCodeLevel.summary, 'Incoming BACS Payment'),
    TypeCode('890', TypeCodeTransaction.misc, TypeCodeLevel.detail, 'Contains Non-monetary Information'),
    TypeCode('906', None, TypeCodeLevel.detail, 'Today’s Opening 1 Day Float'),
    TypeCode('907', None, TypeCodeLevel.detail, 'Today’s Opening 2+ Day Float'),
]
TypeCodes = {
    type_code.code: type_code
    for type_code in TypeCodes
}



#!/usr/bin/env python
import importlib
import os
import sys
import warnings

from setuptools import setup

if sys.version_info[0:2] < (3, 7):
    warnings.warn('This package is tested with Python version 3.7+', stacklevel=1)

root_path = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(root_path, 'README.rst')) as readme:
    README = readme.read()

package_info = importlib.import_module('bai2')

setup(
    name='bai2',
    version=package_info.__version__,
    author=package_info.__author__,
    author_email='dev@digital.justice.gov.uk',
    url='https://github.com/ministryofjustice/bai2',
    packages=['bai2'],
    package_dir={'bai2': 'bai2'},
    include_package_data=True,
    license='MIT',
    description='BAI2 Parser',
    long_description=README,
    keywords='bai2 bookkeeping cash management balance reporting',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    python_requires='>=3.6',  # looser requirement than what's tested
    install_requires=[],
    tests_require=[],
    test_suite='tests',
)



from .models import Record
from .constants import RecordCode


def _build_account_identifier_record(rows):
    fields_str = ''
    for index, row in enumerate(rows):
        field_str = row[1]
        commas_count = field_str.count(',')
        # first row will have account identifier and currency (the common part of the account identifier)
        # and then a multiple of 4 commas (the rest of the account identifier, which are summary components)
        # the rest of the rows will have only the summary components
        summary_commas_count = commas_count - 2 if index == 0 else commas_count

        if field_str:
            if field_str[-1] == '/':
                fields_str += field_str[:-1]
                if summary_commas_count % 4 != 0:
                    # if the number of commas is not a multiple of 4, then we need to add a comma
                    # some banks emit this extra comma, some don't, so we need to normalize it
                    fields_str += ','
            else:
                fields_str += field_str

    fields = fields_str.split(',')
    return Record(code=rows[0][0], fields=fields, rows=rows)


def _build_generic_record(rows):
    fields_str = ''
    for row in rows:
        field_str = row[1]

        if field_str:
            if field_str[-1] == '/':
                fields_str += field_str[:-1] + ','
            else:
                fields_str += field_str + ' '

    fields = fields_str[:-1].split(',')
    return Record(code=rows[0][0], fields=fields, rows=rows)


RecordBuilderFactory = {
    RecordCode.file_header: _build_generic_record,
    RecordCode.group_header: _build_generic_record,
    RecordCode.account_identifier: _build_account_identifier_record,
    RecordCode.transaction_detail: _build_generic_record,
    RecordCode.account_trailer: _build_generic_record,
    RecordCode.group_trailer: _build_generic_record,
    RecordCode.file_trailer: _build_generic_record,
}


def _build_record(rows):
    record_code = rows[0][0]
    return RecordBuilderFactory[record_code](rows)


def record_generator(lines):
    rows = iter(
        [(RecordCode(line[:2]), line[3:]) for line in lines]
    )

    records = [next(rows)]
    while True:
        try:
            row = next(rows)
        except StopIteration:
            break

        if row[0] != RecordCode.continuation:
            yield _build_record(records)
            records = [row]
        else:
            records.append(row)

    yield _build_record(records)


class IteratorHelper:
    def __init__(self, lines):
        self._generator = record_generator(lines)
        self.current_record = None
        self.advance()

    def advance(self):
        self.current_record = next(self._generator)



from collections import OrderedDict
from datetime import date, time
from unittest import TestCase

from bai2 import models, writers, constants


class TransactionDetailWriterTestCase(TestCase):
    def test_transaction_detail_with_no_availability_renders_correctly(self):
        transaction = models.TransactionDetail(
            type_code=constants.TypeCodes['399'],
            amount=2599,
            text='BILLS',
        )

        output = writers.TransactionDetailWriter(transaction).write()
        self.assertEqual(output, ['16,399,2599,,,,BILLS'])

    def test_transaction_detail_with_unknown_availability_renders_correctly(self):
        transaction = models.TransactionDetail(
            type_code=constants.TypeCodes['399'],
            amount=2599,
            funds_type=constants.FundsType.unknown_availability,
            text='BILLS',
        )

        output = writers.TransactionDetailWriter(transaction).write()
        self.assertEqual(output, ['16,399,2599,Z,,,BILLS'])

    def test_transaction_detail_with_text_on_new_line_renders_correctly(self):
        transaction = models.TransactionDetail(
            type_code=constants.TypeCodes['399'],
            amount=2599,
            text='BILLS',
        )

        output = writers.TransactionDetailWriter(
            transaction, text_on_new_line=True
        ).write()
        self.assertEqual(output, ['16,399,2599,,,/', '88,BILLS'])

    def test_transaction_detail_with_immediate_availability_renders_correctly(self):
        transaction = models.TransactionDetail(
            type_code=constants.TypeCodes['399'],
            amount=2599,
            funds_type=constants.FundsType.immediate_availability,
            text='BILLS',
        )

        output = writers.TransactionDetailWriter(transaction).write()
        self.assertEqual(output, ['16,399,2599,0,,,BILLS'])

    def test_transaction_detail_with_distributed_availability_simple_renders_correctly(self):
        transaction = models.TransactionDetail(
            type_code=constants.TypeCodes['399'],
            amount=2599,
            funds_type=constants.FundsType.distributed_availability_simple,
            availability=OrderedDict([('0', 500), ('1', 599), ('>1', 2599)]),
            text='BILLS',
        )

        output = writers.TransactionDetailWriter(transaction).write()
        self.assertEqual(output, ['16,399,2599,S,500,599,2599,,,BILLS'])

    def test_transaction_detail_with_value_dated_availability_renders_correctly(self):
        transaction = models.TransactionDetail(
            type_code=constants.TypeCodes['399'],
            amount=2599,
            funds_type=constants.FundsType.value_dated,
            availability=OrderedDict([('date', date(year=2015, month=10, day=1)),
                                      ('time', None)]),
            text='BILLS',
        )

        output = writers.TransactionDetailWriter(transaction).write()
        self.assertEqual(output, ['16,399,2599,V,151001,,,,BILLS'])

    def test_transaction_detail_with_distributed_availability_renders_correctly(self):
        transaction = models.TransactionDetail(
            type_code=constants.TypeCodes['399'],
            amount=2599,
            funds_type=constants.FundsType.distributed_availability,
            availability=OrderedDict([('1', 500), ('2', 599), ('4', 2599)]),
            text='BILLS'
        )

        output = writers.TransactionDetailWriter(transaction).write()
        self.assertEqual(output, ['16,399,2599,D,3,1,500,2,599,4,2599,,,BILLS'])

    def test_transaction_detail_with_continuation_renders_correctly(self):
        transaction = models.TransactionDetail(
            type_code=constants.TypeCodes['399'],
            amount=2599,
            text='BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS',
        )

        output = writers.TransactionDetailWriter(transaction).write()
        self.assertEqual(
            output,
            ['16,399,2599,,,,BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS',
             '88, BILLS'])


class AccountIdentifierWriterTestCase(TestCase):
    def test_account_identifier_renders_correctly(self):
        account_identifier = models.AccountIdentifier(
            customer_account_number='77777777',
            currency='GBP',
            summary_items=[
                models.Summary(type_code=constants.TypeCodes['010'], amount=10000),
                models.Summary(type_code=constants.TypeCodes['015'], amount=10000),
            ]
        )

        output = writers.AccountIdentifierWriter(account_identifier).write()
        self.assertEqual(
            output,
            ['03,77777777,GBP,010,10000,,,015,10000,,/']
        )

    def test_account_identifier_with_continuation_renders_correctly(self):
        account_identifier = models.AccountIdentifier(
            customer_account_number='77777777',
            currency='GBP',
            summary_items=[
                models.Summary(type_code=constants.TypeCodes['010'],
                               amount=10000),
                models.Summary(type_code=constants.TypeCodes['015'],
                               amount=10000),
                models.Summary(type_code=constants.TypeCodes['045'],
                               amount=10000),
                models.Summary(type_code=constants.TypeCodes['040'],
                               amount=10000),
                models.Summary(type_code=constants.TypeCodes['072'],
                               amount=10000),
                models.Summary(type_code=constants.TypeCodes['074'],
                               amount=10000),
                models.Summary(type_code=constants.TypeCodes['075'],
                               amount=10000),
                models.Summary(type_code=constants.TypeCodes['400'],
                               amount=10000, item_count=175),
                models.Summary(type_code=constants.TypeCodes['100'],
                               amount=10000, item_count=50),
            ]
        )

        output = writers.AccountIdentifierWriter(account_identifier).write()
        self.assertEqual(
            output,
            ['03,77777777,GBP,010,10000,,,015,10000,,,045,10000,,,040,10000,,,072,10000,,/',
             '88,074,10000,,,075,10000,,,400,10000,175,,100,10000,50,/']
        )

    def test_account_identifier_with_summary_availability_renders_correctly(self):
        account_identifier = models.AccountIdentifier(
            customer_account_number='77777777',
            currency='GBP',
            summary_items=[
                models.Summary(
                    type_code=constants.TypeCodes['010'],
                    amount=10000,
                    funds_type=constants.FundsType.distributed_availability_simple,
                    availability=OrderedDict([('0', 100), ('1', 200), ('>1', 300)])
                ),
                models.Summary(
                    type_code=constants.TypeCodes['015'],
                    amount=10000,
                    funds_type=constants.FundsType.distributed_availability_simple,
                    availability=OrderedDict([('0', 100), ('1', 200), ('>1', 300)])
                )
            ]
        )

        output = writers.AccountIdentifierWriter(account_identifier).write()
        self.assertEqual(
            output,
            ['03,77777777,GBP,010,10000,,S,100,200,300,015,10000,,S,100,200,300/']
        )


class AccountTrailerWriterTestCase(TestCase):
    def test_account_trailer_renders_correctly(self):
        account_trailer = models.AccountTrailer(
            account_control_total=100,
            number_of_records=4
        )

        output = writers.AccountTrailerWriter(account_trailer).write()
        self.assertEqual(
            output,
            ['49,100,4/']
        )


class GroupHeaderWriterTestTcase(TestCase):
    def test_group_header_renders_correctly(self):
        group_header = models.GroupHeader(
            ultimate_receiver_id='8888888',
            originator_id='CITIGB00',
            group_status=constants.GroupStatus.update,
            as_of_date=date(year=2015, month=7, day=15),
            as_of_time=time(hour=23, minute=40),
            currency='GBP',
            as_of_date_modifier=constants.AsOfDateModifier.final_previous_day
        )

        output = writers.GroupHeaderWriter(group_header).write()
        self.assertEqual(
            output,
            ['02,8888888,CITIGB00,1,150715,2340,GBP,2/']
        )


class GroupTrailerWriterTestCase(TestCase):
    def test_account_trailer_renders_correctly(self):
        group_trailer = models.GroupTrailer(
            group_control_total=100,
            number_of_accounts=1,
            number_of_records=6
        )

        output = writers.GroupTrailerWriter(group_trailer).write()
        self.assertEqual(
            output,
            ['98,100,1,6/']
        )


class Bai2FileHeaderWriterTestCase(TestCase):
    def test_file_header_renders_correctly(self):
        file_header = models.Bai2FileHeader(
            sender_id='CITIDIRECT',
            receiver_id='8888888',
            creation_date=date(year=2015, month=7, day=15),
            creation_time=time(hour=23, minute=40),
            file_id='00131100',
            physical_record_length=None,
            block_size=None,
            version_number=2
        )

        output = writers.Bai2FileHeaderWriter(file_header).write()
        self.assertEqual(
            output,
            ['01,CITIDIRECT,8888888,150715,2340,00131100,,,2/']
        )


class Bai2FileTrailerWriterTestCase(TestCase):
    def test_file_trailer_renders_correctly(self):
        file_trailer = models.Bai2FileTrailer(
            file_control_total=100,
            number_of_groups=1,
            number_of_records=8
        )

        output = writers.Bai2FileTrailerWriter(file_trailer).write()
        self.assertEqual(
            output,
            ['99,100,1,8/']
        )


class AccountWriterTestCase(TestCase):
    @staticmethod
    def create_account_section():
        transactions = []
        transactions.append(models.TransactionDetail(
            type_code=constants.TypeCodes['399'],
            amount=2599,
            text='BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS',
        ))
        transactions.append(models.TransactionDetail(
            type_code=constants.TypeCodes['399'],
            amount=1000,
            funds_type=constants.FundsType.immediate_availability,
            text='OTHER',
        ))

        account_identifier = models.AccountIdentifier(
            customer_account_number='77777777',
            currency='GBP',
            summary_items=[
                models.Summary(type_code=constants.TypeCodes['010'], amount=10000),
                models.Summary(type_code=constants.TypeCodes['015'], amount=10000),
            ]
        )

        return models.Account(header=account_identifier, children=transactions)

    def test_account_renders_correctly(self):
        account = AccountWriterTestCase.create_account_section()

        output = writers.AccountWriter(account).write()
        self.assertEqual(
            output,
            [
                '03,77777777,GBP,010,10000,,,015,10000,,/',
                '16,399,2599,,,,BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS',
                '88, BILLS',
                '16,399,1000,0,,,OTHER',
                '49,23599,5/',
            ]
        )


class GroupWriterTestCase(TestCase):
    @staticmethod
    def create_group_section():
        accounts = []
        accounts.append(AccountWriterTestCase.create_account_section())
        accounts.append(AccountWriterTestCase.create_account_section())

        group_header = models.GroupHeader(
            ultimate_receiver_id='8888888',
            originator_id='CITIGB00',
            group_status=constants.GroupStatus.update,
            as_of_date=date(year=2015, month=7, day=15),
            as_of_time=time(hour=23, minute=40),
            currency='GBP',
            as_of_date_modifier=constants.AsOfDateModifier.final_previous_day
        )

        return models.Group(header=group_header, children=accounts)

    def test_group_renders_correctly(self):
        group = GroupWriterTestCase.create_group_section()

        output = writers.GroupWriter(group).write()
        self.assertEqual(
            output,
            [
                '02,8888888,CITIGB00,1,150715,2340,GBP,2/',
                '03,77777777,GBP,010,10000,,,015,10000,,/',
                '16,399,2599,,,,BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS',
                '88, BILLS',
                '16,399,1000,0,,,OTHER',
                '49,23599,5/',
                '03,77777777,GBP,010,10000,,,015,10000,,/',
                '16,399,2599,,,,BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS',
                '88, BILLS',
                '16,399,1000,0,,,OTHER',
                '49,23599,5/',
                '98,47198,2,12/',
            ]
        )


class Bai2FileWriterTestCase(TestCase):
    @staticmethod
    def create_bai2_file():
        groups = []
        groups.append(GroupWriterTestCase.create_group_section())
        groups.append(GroupWriterTestCase.create_group_section())

        file_header = models.Bai2FileHeader(
            sender_id='CITIDIRECT',
            receiver_id='8888888',
            creation_date=date(year=2015, month=7, day=15),
            creation_time=time(hour=23, minute=40),
            file_id='00131100',
            physical_record_length=None,
            block_size=None,
            version_number=2
        )

        return models.Bai2File(header=file_header, children=groups)

    def test_bai2_file_renders_correctly(self):
        bai2_file = Bai2FileWriterTestCase.create_bai2_file()

        output = writers.Bai2FileWriter(bai2_file).write()
        self.assertEqual(
            output,
            [
                '01,CITIDIRECT,8888888,150715,2340,00131100,,,2/',
                '02,8888888,CITIGB00,1,150715,2340,GBP,2/',
                '03,77777777,GBP,010,10000,,,015,10000,,/',
                '16,399,2599,,,,BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS',
                '88, BILLS',
                '16,399,1000,0,,,OTHER',
                '49,23599,5/',
                '03,77777777,GBP,010,10000,,,015,10000,,/',
                '16,399,2599,,,,BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS',
                '88, BILLS',
                '16,399,1000,0,,,OTHER',
                '49,23599,5/',
                '98,47198,2,12/',
                '02,8888888,CITIGB00,1,150715,2340,GBP,2/',
                '03,77777777,GBP,010,10000,,,015,10000,,/',
                '16,399,2599,,,,BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS',
                '88, BILLS',
                '16,399,1000,0,,,OTHER',
                '49,23599,5/',
                '03,77777777,GBP,010,10000,,,015,10000,,/',
                '16,399,2599,,,,BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS BILLS',
                '88, BILLS',
                '16,399,1000,0,,,OTHER',
                '49,23599,5/',
                '98,47198,2,12/',
                '99,94396,2,26/',
            ]
        )
